#!/usr/bin/perl
use strict;
use warnings;
use Gtk2 '-init';
use Gtk2::SimpleMenu;
use Gtk2::Pango;
use Gtk2::Helper;
use feature ":5.10";
use Bio::SearchIO;

#Global variables
my $seqview;
my $db_table;
my $output;
my $file_box;
my $all;
my $assess_box;
my $program_box;
my $cpu_box;
my $load_button;
my $run_button;
my %buttons;
my $file;
my $pipe;
my $pid;

my $num_processors = 1;
my $LARGE_QUERY = 10000;
use constant TRUE  => 1;
use constant FALSE => 0;

#Construct GUI
my $window = Gtk2::Window->new;
$window -> set_title('CBOrg v1.0');
$window -> signal_connect(delete_event => \&delete_dialog);
$window -> set_position('center');
$window -> set_default_size(900, 700);
$window -> set_border_width(3);

my $vbox = Gtk2::VBox -> new();

my $menu_tree = create_menu_tree();
my $menu = Gtk2::SimpleMenu -> new(
                menu_tree               => $menu_tree,
                default_callback        => \&default_callback,
            );

my $frame = Gtk2::Frame -> new();
$frame -> set_shadow_type('in');

$vbox -> pack_start($menu -> {widget}, FALSE, FALSE, 0);
$vbox -> pack_start($frame, TRUE, TRUE, 0);

my $opt_frame = Gtk2::Frame -> new("Options");
$opt_frame -> set_shadow_type('out');

my $opt_vbox = Gtk2::VBox -> new();
$opt_frame -> add($opt_vbox);

my $table = create_main_table($opt_frame);

$frame -> add($table);

my $blast_table = create_blast_table();
$db_table = create_db_table();

my $db_frame = Gtk2::Frame -> new("Databases");
$db_frame -> add($db_table);
$db_frame -> set_label_align(0.5, 0.5);
$db_frame -> set_shadow_type('in');

my $blast_frame = Gtk2::Frame -> new("BLAST Parameters");
$blast_frame -> add($blast_table);
$blast_frame -> set_label_align(0.5, 0.5);
$blast_frame -> set_shadow_type('in');

$opt_vbox -> pack_start($blast_frame, TRUE, TRUE, 0);
$opt_vbox -> pack_start($db_frame, TRUE, TRUE, 0);

$window -> add($vbox);
$frame -> set_size_request (0, 0);

#Run GUI
$window -> show_all;
Gtk2 -> main;



################################################################################
#                               Subroutines                                    #
################################################################################

sub delete_dialog{
    my $dialog = Gtk2::MessageDialog -> new($window, 'destroy-with-parent',
                                            'question', 'yes-no',
                                            "Are you sure you want to exit?\n
                                            All unsaved changes will be lost"
                                            );
    my $response = $dialog->run;
    $dialog -> destroy;
    if($response eq 'yes'){
        Gtk2 -> main_quit;
    }
    1;
}

sub open_document {
	my $start = shift;

	my $file_dialog = Gtk2::FileChooserDialog->new ('Choose a Starting Folder',
                                                    $window, 'open',
                                                    'gtk-cancel'	=> 'cancel',
                                                    'gtk-ok'	=> 'ok');
    
	$file_dialog->set_select_multiple (FALSE);

	my $response = $file_dialog->run;
	$file = $file_dialog->get_filename;
	$file_dialog->destroy;

	if ($response eq 'cancel') {
		return FALSE;
	} else {
        loadFile();
		return $file;
	}
}

sub loadFile {
	my $sep = $/;
	$/ = undef;
	
	open(IN, "$file") or die "Could not open $file. Exiting...\n";
	my $data = <IN>;
	close(IN);
	
	$/ = $sep;
	
	my $buffer = $seqview -> get_buffer();
	my $start_iter = $buffer -> get_start_iter();
	my $end_iter = $buffer -> get_end_iter();
	$buffer -> delete($start_iter, $end_iter);
	$start_iter = $buffer -> get_start_iter();
	$buffer -> insert($start_iter, $data);
}

sub create_menu_tree {
    my $menu_tree = [
        _File => {
                item_type => '<Branch>',
                children => [
                        _Open => {
                                item_type       => '<StockItem>',
                                callback        => \&open_document,
                                accelerator     => '<ctrl>O',
                                extra_data      => 'gtk-open',
                        },
                        _Quit => {
                                item_type       => '<StockItem>',
                                callback        => \&delete_dialog,
                                accelerator     => '<ctrl>Q',
                                extra_data      => 'gtk-quit',
                        },
                ],
        },
        _Help => {
                item_type => '<LastBranch>',
                children => [
                        _Help => {
                                item_type       => '<StockItem>',
                                callback_action => 3,
                                accelerator     => '<ctrl>H',
                                extra_data      => 'gtk-help',
                        }
                ],
        },
    ];
    
    return $menu_tree;
}

sub create_main_table{
    my $opt_frame = shift;
    
    my $main = Gtk2::Table -> new(4, 4, FALSE);
    
    my $seq_scroll = Gtk2::ScrolledWindow -> new(undef, undef);
    $seq_scroll -> set_shadow_type('in');
    $seq_scroll -> set_policy('automatic', 'automatic');
    $seq_scroll -> set_size_request(300, 200);
    $seq_scroll -> set_border_width(5);
    
    my $out_scroll = Gtk2::ScrolledWindow -> new(undef, undef);
    $out_scroll -> set_shadow_type('in');
    $out_scroll -> set_policy('automatic', 'automatic');
    $out_scroll -> set_size_request(300, 200);
    $out_scroll -> set_border_width(5);
    
    $seqview = Gtk2::TextView -> new();
    $output = Gtk2::TextView -> new();
    $file_box = Gtk2::Entry -> new();
    
    $seqview -> set_wrap_mode('char');
    $output -> set_wrap_mode('word');
    
    $seq_scroll -> add($seqview);
    $out_scroll -> add($output);
    
    $load_button = Gtk2::Button -> new('Load');
    my $stop_button = Gtk2::Button -> new('Stop');
    $run_button = Gtk2::Button -> new('Run');
    
    $load_button -> signal_connect('clicked' => \&open_document);
    $stop_button -> signal_connect('clicked' => \&halt);
    $run_button -> signal_connect('clicked' => \&run_blast);

    $main -> attach($seq_scroll, 0, 2, 0, 1,
                     ['fill'],['expand', 'fill'], 10, 5 );
    $main -> attach($out_scroll, 2, 4, 0, 2,
                     ['fill'],['expand', 'fill'], 10, 5 );
    $main -> attach($file_box, 0, 1, 1, 2,
                     ['fill'], ['fill'], 5, 5);
    $main -> attach($load_button, 1, 2, 1, 2,
                     ['fill'], ['fill'], 5, 5);
    $main -> attach($opt_frame, 0, 4, 2, 3,
                     ['expand', 'fill'], ['fill'], 5, 5);
    $main -> attach($stop_button, 1, 2, 3, 4,
                     ['fill'], ['fill'], 5, 5);
    $main -> attach($run_button, 2, 3, 3, 4,
                     ['fill'], ['fill'], 5, 5);
    
    return $main;
}

sub create_db_table{
    my @dbs = get_dbs();
    my $num_rows = scalar(@dbs);
    $num_rows++;
    my $num_columns = 4;
    
    my $database_table = Gtk2::Table -> new($num_rows, $num_columns, FALSE);
    
    $all = Gtk2::CheckButton -> new('All');
    $all -> signal_connect(toggled => \&check_all);
    
    $database_table -> attach($all, 0, 1, 0, 1,
                  ['fill'], ['fill'], 5, 5);
    
    my $prot_label = Gtk2::Label -> new('Proteome');
    $prot_label -> set_alignment(0, 0.5);
    $database_table -> attach($prot_label, 1, 2, 0, 1,
                  ['expand', 'fill'], ['fill'], 5, 5);
    
    my $mito_label = Gtk2::Label -> new('Mito-Proteome');
    $mito_label -> set_alignment(0, 0.5);
    $database_table -> attach($mito_label, 2, 3, 0, 1,
                  ['expand', 'fill'], ['fill'], 5, 5);
    
    my $chloro_label = Gtk2::Label -> new('Chloro-Proteome');
    $chloro_label -> set_alignment(0, 0.5);
    $database_table -> attach(Gtk2::Label -> new('Chloro-Proteome'), 3, 4, 0, 1,
                  ['expand', 'fill'], ['fill'], 5, 5);
    
    for(my $i = 1; $i < $num_rows; $i++){
        my $entry = $i - 1;
        my $temp_label = Gtk2::Label -> new("$dbs[$entry]");
        $temp_label -> set_alignment(0, 0.5);
        $database_table -> attach($temp_label, 0, 1, $i, $i + 1,
                      ['expand', 'fill'], ['expand', 'fill'], 5, 5);
        
        my @temp;
        for(my $x = 1; $x < $num_columns; $x++){
            my $temp = Gtk2::CheckButton -> new();
            $database_table -> attach($temp, $x, $x + 1, $i, $i + 1,
                          ['fill'], ['fill'], 5, 2);
            push @temp, $temp;
        }
        $buttons{$dbs[$entry]} = [@temp];
    }
    
    return $database_table;
}

sub create_blast_table{
	my $blast = Gtk2::Table -> new(6, 4, FALSE);
	
	$program_box = Gtk2::ComboBox -> new_text();
	$program_box -> append_text("tblastx");
	$program_box -> append_text("blastx");
    $program_box -> append_text("blastp");
	$program_box -> set_title("Program");
	$program_box -> set_active(1);
	
	$assess_box = Gtk2::ComboBox -> new_text();
	$assess_box -> append_text("e-value");
	$assess_box -> append_text("Raw Score");
	$assess_box -> set_title("Assessment");
	$assess_box -> set_active(1);
    
    $cpu_box = Gtk2::ComboBox -> new_text();
    $cpu_box -> append_text("1");
    $cpu_box -> append_text("2");
    $cpu_box -> append_text("3");
    $cpu_box -> append_text("4");
    $cpu_box -> append_text("5");
    $cpu_box -> append_text("6");
    $cpu_box -> append_text("7");
    $cpu_box -> append_text("8");
    $cpu_box -> set_title("CPU");
    $cpu_box -> set_active(0);
	
	$blast -> attach(Gtk2::Label -> new("Program: "), 0, 1, 0, 1,
					 ['fill'], ['fill'], 5, 3);
	$blast -> attach($program_box, 1, 2, 0, 1,
					 ['expand', 'fill'], ['expand', 'fill'], 5, 5);
	$blast -> attach(Gtk2::Label -> new("Assessment: "), 2, 3, 0, 1,
					 ['fill'], ['fill'], 5, 3);
	$blast -> attach($assess_box, 3, 4, 0, 1,
					 ['expand', 'fill'], ['expand', 'fill'], 5, 5);
    $blast -> attach(Gtk2::Label -> new("# of CPUs: "), 4, 5, 0, 1,
					 ['fill'], ['fill'], 5, 3);
    $blast -> attach($cpu_box, 5, 6, 0, 1,
                     ['expand', 'fill'], ['expand', 'fill'], 5, 5);
	
	return $blast;
}

sub get_dbs{
    my $dbfile = "databases.txt";
    open(IN, "$dbfile") or die "Could not open $file. Exiting...\n";
    my @temp = <IN>;
    close(IN);
    
    my @dbs;
    foreach(@temp){
        my $start = substr($_,0,1);
        if($start ne '#'){
            if($_ =~ /[a-z]/i){
                my @temp = split /:/, $_;
                push @dbs, $temp[0];   
            }
        }
    }
    return @dbs;
}

sub get_blast_dbs{
    my $dbfile = "databases.txt";
    open(IN, "$dbfile") or die "Could not open $file. Exiting...\n";
    my @temp = <IN>;
    close(IN);
    
    my %dbs;
    
    foreach(@temp){
        my $start = substr($_,0,1);
        if($start ne '#'){
            if($_ =~ /[a-z]/i){
                $_ =~ s/[\r\n]+//g;
                my @temp = split /: /, $_;
                my $org = $temp[0];
                my $data = $temp[1];
                my @databases = split / /, $data;
                
                $dbs{$org} = [@databases];
            }
        }
    }
    
    return %dbs;
}

sub check_all{
    my $checked = $all -> get_active();
    
	foreach my $org (keys(%buttons)){
        my @temp = @{$buttons{$org}};
        foreach my $db (@temp){
            if(!$checked){
                my $active = $db -> get_active();
                if($active){
                    $db -> set_active(FALSE); 
                }
            }else{
                my $active = $db -> get_active();
                if(!$active){
                    $db -> set_active(TRUE); 
                }
            }
        }
    }
}

sub run_blast{
    my $selected = $program_box -> get_active();
    my $proc_select = $cpu_box -> get_active();
    my $algorithm;
    
    setSensitivity(FALSE);
    
    given($selected){
        when("0"){ $algorithm = "tblastx"; }
        when("1"){ $algorithm = "blastx"; }
        when("2"){ $algorithm = "blastp"; }
    }
    
    $num_processors = $proc_select + 1;
    
    my %dbs = get_blast_dbs();
    
    my $buffer = $output -> get_buffer();
    my $end_iter = $buffer -> get_end_iter();
    
    my @blast_files;
    my @mito_files;
    my @subtract_files;
    my @chloro_files;
        
	foreach my $org (keys(%buttons)){
        my @temp = @{$buttons{$org}};
        my $i = 0;
        foreach my $db (@temp){
            my $active = $db -> get_active();
            if($active){
                my $database = $dbs{$org}[$i];
                if($database ne 'F'){
                    my $out = "$database" . ".blast";
                    $out =~ s/^.*\/([^\/]*)/$1/;
                    
                    my $main_loop = Glib::MainLoop -> new();
                    my $execute_command = "blastall -p $algorithm -a $num_processors -b 5 -v 5 -z 20000000 -F " .
                                            '"m S"' . " -i $file -d $database -o $out";
                    $pid = open CHILD, "$execute_command |" or die "Can't fork BLAST: $!";
                    
                    printToBuffer("BLAST executing with PID: $pid\n");
                    
                    my $data = "Executing: blastall -p $algorithm -a $num_processors -b 5 -v 5 -z 20000000 -F " .
                               '"m S"' . " -i $file -d $database -o $out\n\n";
                    printToBuffer($data);   
                    
                    Glib::IO -> add_watch(fileno CHILD, ['hup'],
								  sub {
                                    $main_loop->quit;
                                    FALSE;
								  }
								 );
					
                    $main_loop -> run();
					
                    close CHILD or warn "$pid died with exit status ". ($? >> 8)."\n";
                    if(($? >> 8) != 0){
                        printToBuffer("Exiting BLAST subroutine...\n");
                        
                        my $exit_message = "$pid died with exit status ". ($? >> 8)."\n";
                        printToBuffer($exit_message);
                        
                        return 1;
                    }
                    if($i == 0){
                        push @subtract_files, $out;
                    }elsif($i == 1){
                        push @mito_files, $out;
                    }elsif($i == 2){
                        push @chloro_files, $out;
                    }else{
                        printToBuffer("Warning: Too Many database types!\n");
                        die();
                    }
                }else{
                    if($i == 0){
                        push @subtract_files, "F";
                    }elsif($i == 1){
                        push @mito_files, "F";
                    }elsif($i == 2){
                        push @chloro_files, "F";
                    }else{
                        printToBuffer("Warning: Too Many database types!\n");
                        die();
                    } 
                }
            }
            $i++;
        }
    }
    
    push @blast_files, [@subtract_files];
    push @blast_files, [@mito_files];
    push @blast_files, [@chloro_files];
    
    printToBuffer("BLAST runs finished. Now performing comparisons...\n");
    
    my $main_loop = Glib::MainLoop -> new();
    blastCompare(@blast_files); 
    Glib::IO -> add_watch(fileno CHILD, ['hup'],
						sub {
                            $main_loop->quit;
                            FALSE;
							}
						);
    $main_loop -> run();
}

sub blastCompare{
    my @files = @_;
    my @outfiles_db;
    
    my $outroot = $file;
    $outroot =~ s/^.*\/([^\/]*)/$1/;
    
    my $assessment = $assess_box -> get_active();
    
    printToBuffer("Now parsing options and running BLAST comparisons\n");
    
    my @subtr_db = @{$files[0]};
	my @mito_db = @{$files[1]};
	my @chloro_db = @{$files[2]};

	my @num;
	$num[0] = scalar(@subtr_db);
	$num[1] = scalar(@mito_db);
	$num[2] = scalar(@chloro_db);
    
    my $num_org = $num[0];
    
    for(my $i = 0; $i < 3; $i++){
        if($num[$i]){
            if($num[$i] != $num_org){
                if($num_org != 0){
                    printToBuffer("Unequal number of files being compared between databases. Exiting...\n");
                    halt();
                }
            }
        }
    }
    
	for(my $i = 0; $i < $num_org; $i++){
		my $mito_report;
		my $chloro_report;
		my $subtr_report;
		
		my @reports = [];
		my @dbs = [];
		my $y = 0;
		
		if($num[0]){
			if($subtr_db[$i] ne "F"){
                printToBuffer("Retrieving subtractive database (cytosolic) results\n");
                ($reports[$y], $dbs[$y]) = getReport($subtr_db[$i], "S");
				$y++;
			}else{
                printToBuffer("Placeholder in place for subtractive blast\n");
			}
		}
        if($num[1]){
			if($mito_db[$i] ne "F"){
                printToBuffer("Retrieving mitoblast results\n");
                ($reports[$y], $dbs[$y]) = getReport($mito_db[$i], "M");
                $y++;
			}else{
                printToBuffer("Placeholder in place for mito blast\n");
			}
			
		}
		if($num[2]){
			if($chloro_db[$i] ne "F"){
                printToBuffer("Retrieving chloroblast results\n");
                ($reports[$y], $dbs[$y]) = getReport($chloro_db[$i], "C");
                $y++;
			}else{
                printToBuffer("Placeholder in place for chloro blast\n");
			}
		}
		
		my $outfile = $outroot . "-$i";
        push @outfiles_db, $outfile;
        
		open(OUT, ">$outfile") or die "Could not create $output or open for appending!\n"; 
		print OUT "Files being compared: ";
        
        if($subtr_db[$i]){
			print OUT "$subtr_db[$i], ";
		}
        
		if($mito_db[$i]){
			print OUT "$mito_db[$i], ";
		}
        
		if($chloro_db[$i]){
			print OUT "$chloro_db[$i], ";
		}
        
		print OUT "\n";
	
		my $num_reports = scalar(@reports);
		my $not_done = 1;
        
        printToBuffer("Parsing BLAST results\n");
		
		my $w = 1;
		while(my $result1 = $reports[0] -> next_result){
		
			my $x = 0;
			my @results = [];
			my @hits = [];
			my @scores = [];
			my @expect = [];
			my @dbs_copy = [];
			my $best_index = 0;
			my $query = "";
			my $query_desc = "";
			my $query_length = 0;
			my $algorithm_str = $result1 -> algorithm;
			my $algorithm_ver = $result1 -> algorithm_version;
			
			if($w == 1){
				print OUT "Algorithm Version: $algorithm_ver\n";
				print OUT "Algorithm String: $algorithm_str\n";
			}
			
			#Need to decide which score is the best and print out all relevant information to the file
			for($x = 0; $x < $num_reports; $x++){
				
				if($x == 0){
					$results[$x] = $result1;
				}else{
					eval{
						$results[$x] = $reports[$x] -> next_result;
					};
					if($@){
						$@ = "";
						last;
					}
				}
				
				eval{
					$query = $results[0] -> query_name;
					$query_desc = $results[0] -> query_description;
					$query_length = $results[0] -> query_length;
					$hits[$x] = $results[$x] -> next_hit;
				};
                
				if($@){
					$not_done = 0;
					$@ = "";
					last;
				}
				
				eval{
					#$scores[$x] = $hits[$x] -> raw_score;
					$scores[$x] = $hits[$x] -> bits;
				};
				if($@){
					$scores[$x] = 0;
					$@ = "";
				}
				
				eval{
					$expect[$x] = $hits[$x] -> significance;
				};
				if($@){
					$expect[$x] = 0;
					$@ = "";
				}
			}
				
			#sort all of the necessary arrays based on the value in the score array
			#appears that the only arrays that need to be sorted are the scores, e-value, dbs, and hits
			#so that they are all in the same order. This sort uses a simple bubble sort algorithm
				
			my $OL;
			my $IL;
			my $temp_score = 0;
			my $temp_expect = 0;
			my $temp_dbs = "";
			my $temp_hits = "";
			
			@dbs_copy = @dbs;
				
			for($OL = 0; $OL < scalar(@scores); $OL++){
				for($IL = $OL; $IL < scalar(@scores); $IL++){
					if($scores[$IL] > $scores[$OL]){
							
						$temp_score = $scores[$IL];
						$scores[$IL] = $scores[$OL];
						$scores[$OL] = $temp_score;
						
						$temp_dbs = $dbs_copy[$IL];
						$dbs_copy[$IL] = $dbs_copy[$OL];
						$dbs_copy[$OL] = $temp_dbs;
							
						$temp_hits = $hits[$IL];
						$hits[$IL] = $hits[$OL];
						$hits[$OL] = $temp_hits;
						
						$temp_expect = $expect[$IL];
						$expect[$IL] = $expect[$OL];
						$expect[$OL] = $temp_expect;
					}
				}
			}
			
			if($scores[0]){
				if($query_length >= $LARGE_QUERY){
					print OUT "*LARGE QUERY*\t";
				}else{
					print OUT " \t";
				}
				
				#generate the ranges for query and hit based on 1st ranked HSP
				my @query_positions = [];
				my @hit_positions = [];
				my $hsp_rank = "";
				
				while(my $hsp = $hits[0] -> next_hsp){
					$hsp_rank = "";
					$hsp_rank = $hsp -> rank;
					
					if($hsp_rank == 1){
						@query_positions = [];
						@hit_positions = [];
						@query_positions = $hsp -> range('query');
						@hit_positions = $hsp -> range('hit');
					}
				}
				print OUT "$dbs_copy[0] score: $scores[0] (E: $expect[0]), $dbs_copy[1] score: $scores[1] (E: $expect[1])\t Query: $query Description: $query_desc\t Alignment Info(Query/Hit) - Start: $query_positions[0]/$hit_positions[0] End: $query_positions[1]/$hit_positions[1]\t" . "Hit: " . $hits[0] -> name . " " . $hits[0] -> description . "\t\n";
			}else{
				print OUT " \tNO HITS\t Query: $query, Description: $query_desc\n";	
			}
			$w++;
		}
		close(OUT);
		my $set_file = $i + 1;
        printToBuffer("Done the $set_file set of files... starting next set if any\n");
	} 
    
    printToBuffer("Done generating blast comparison files, Now compiling data and comparing results\n");
	
	my @data_2db;
	my @temp_db;
	my $max_lines = 0;
	my $temp_lines = 0;

	for(my $i = 0; $i < scalar(@outfiles_db); $i++){
		open(IN, "$outfiles_db[$i]") or die "Could not open $outfiles_db[$i] for reading!\n";
		@temp_db = <IN>;
		close(IN);

		$temp_lines = scalar(@temp_db);
		if($i == 0){
			$max_lines = $temp_lines;
		}else{
			if($temp_lines > $max_lines){
				printToBuffer("WARNING: Files are not the same length!\n");
				printToBuffer("Exiting...\n");
				exit();
			}
		}
		push @data_2db, [@temp_db];
	}

	my $merged = $outroot . "-merged";
	my $j;
	
	my $inner = scalar(@data_2db);
	
	open(MERGED, ">$merged") or die "Could not create $merged!\n";
	for($j = 0; $j < $max_lines; $j++){
		for(my $i = 0; $i < $inner; $i++){
			print MERGED "$data_2db[$i][$j]";
		}
		print MERGED "-" x 50;
		print MERGED "\n";
	}
	close(MERGED);
	
	printToBuffer("BLAST comparisons finished. Sorting output...\n");
	findMito($merged, $num_org);
}

sub findMito{
    my $input = shift;
    my $num_organisms = shift;
    my $mito_flag = "M";
    
    my $outfile = $input;
    $outfile =~ s/-merged//;
    
    my $no_hit_out = $outfile;
    my $non_mito_out = $outfile;
    my $mito_sequences_out = $outfile;
    my $non_mito_sequences_out = $outfile;
    my $no_hit_sequences_out = $outfile;
    
    $outfile = $outfile . "-putative-mito";
    $no_hit_out = $no_hit_out . "-no-hit";
    $non_mito_out = $non_mito_out . "-non-mito";
    $mito_sequences_out = $mito_sequences_out . "-mito-sequences.fasta";
    $non_mito_sequences_out = $non_mito_sequences_out . "-non-mito-sequences.fasta";
    $no_hit_sequences_out = $no_hit_sequences_out . "-no-hit-sequences.fasta";
    
    my $sep = $/;
    $/ = undef;

    open(IN, "$input") or die "Could not open $input for reading, please check that file exists...\nExiting...\n";
    my $raw_merged = <IN>;	
    close(IN);
    
    my @merged = split /--------------------------------------------------\n/, $raw_merged;
    my $files = shift @merged;
    my $algorithm_ver = shift @merged;
    my $algorithm_str = shift @merged;
    
    my @temp;
    my @temp2;
    my @query_list;
    my $x = 0;
    
    foreach(@merged){
    	@temp = [];
    	@temp2 = [];
    	@temp = split /\n/, $_;
    	foreach(@temp){
    		@temp2 = split /\t/, $_;
    		$query_list[$x] = $temp2[1];
    		$x++;
    	}
    }
    
    #the number of organisms gives us how many elements of the array should
    #be grouped together. If n = 2 than elements 0 and 1, 2 and 3, etc go together.
    #If n = 2 than 0, 1 and 2 go together, etc. Loop through the master array and
    #temporarily assign the appropriate number of elements to a new array to work
    #with
    
    for(my $i = 0; $i < scalar(@query_list); $i = $i + $num_organisms){
    	my @query = [];
    	for(my $y = 0; $y < $num_organisms; $y++){
    		my $index = $i + $y;
    		$query[$y] = $query_list[$index];
    	}
    	
    	#Sort into at least one mito and all agree no hit categories
    	#To link back to @merged divide @query_list by $num_organisms
    	#if there are two organisms than $merged[0] contains two sets of scores
    	#and these scores are held in $query_list[0] and $query_list[1]
    	#Therefore $i/$num_organisms is the index appropriate in @merged
    	
    	my @db_hits;
    	foreach(@query){
    		my $hit = substr($_, 0, 1);
    		push @db_hits, $hit;
    	}
    	
    	#Now go through and compare the database type that is labelled. 'N' No Hit
    	my $db = $db_hits[0];
    	my $agree_no_hit = 0;
        my $agree_mito = 0;
        if($db eq 'M'){
            $agree_mito = 1;
        }else{
            foreach(@db_hits){
                if($_ eq 'M'){
                    $agree_mito = 1;
                    last();
                }
                if($_ eq 'N'){
                    $agree_no_hit++;
                }
            }
    	}
        
        my $master_index = $i/$num_organisms;
        if($agree_no_hit >= $num_organisms){
            printFile($merged[$master_index],$no_hit_out);
            printFasta($merged[$master_index], $no_hit_sequences_out);
        }elsif($agree_mito){
            printFile($merged[$master_index],$outfile);
            printFasta($merged[$master_index], $mito_sequences_out);
        }else{
            printFile($merged[$master_index],$non_mito_out);
            printFasta($merged[$master_index], $non_mito_sequences_out);
        }
    }
    
    printToBuffer("Sorting complete...\n\n");
    setSensitivity(TRUE);
}

sub printToBuffer{
    my $message = shift;
    my $buffer = $output -> get_buffer();
    
    my $end_iter = $buffer -> get_end_iter();
    $buffer -> insert($end_iter, $message);
    my $end_mark = $buffer -> create_mark ('end', $buffer->get_end_iter, FALSE);
    $output -> scroll_to_mark ($end_mark, 0.0, TRUE, 0.0, 1.0);
}

sub getReport{
    my ($file, $db) = @_;
    my $report = new Bio::SearchIO(-format => 'blast', -file => $file);
    
    return ($report, $db);
}

sub printFasta{
    my ($line, $outfile) = @_;
    my @sequences = getSequences();
    
    foreach (@sequences){
        my @temp = split /\t/, $line;
        my $id = $temp[2];
        $id =~ s/Query: (.+) Description: .*/$1/;
        $id =~ s/\ +$//;
        $id =~ s/^\ +//;
        my @sequence = split /[\r\n]+/, $_;
        $id =~ s/\cn//gs;
        $sequence[0] =~ s/\cn//gs;
        if(/$id/){
            open(OUT, ">>$outfile") or die "Could not open $outfile for appending!\n";
            print OUT $_;
            close(OUT);
            last();
        }
    }
}

sub printFile{
    my($line, $outfile) = @_;
    open(OUT, ">>$outfile") or die "Could not open $outfile for appending!\n";
    print OUT "-" x 50;
    print OUT "\n$line";
    close(OUT);
}

sub getSequences{
    my $sep = $/;
    $/ = ">";
    open(IN, "$file") or die "Could not read $file!\n";
    my @sequences = <IN>;
    close(IN);
    $/ = $sep;
    shift @sequences;
    
    return @sequences;
}

sub setSensitivity{
    my $constant = shift;
    
    $program_box -> set_sensitive($constant);
    $cpu_box     -> set_sensitive($constant);
    $assess_box  -> set_sensitive($constant);
    $file_box    -> set_sensitive($constant);
    $seqview     -> set_sensitive($constant);
    $db_table    -> set_sensitive($constant);
    $load_button -> set_sensitive($constant);
    $run_button  -> set_sensitive($constant);
}

sub halt{
    printToBuffer("Killing process $pid...\n\n");
	`kill -9 $pid`;
}

sub default_callback{
    my (undef, $callback_action, $menu_item) = @_;
    print "callback action number $callback_action\n";
}